import pytest
from django.utils import timezone
from rest_framework.test import APIRequestFactory
from wbcore.contrib.authentication.factories import UserFactory
from wbcore.contrib.directory.factories.entries import PersonWithEmployerFactory
from wbcore.contrib.directory.models import Company, Person

from wbcrm.filters.signals import choice_noactivity
from wbcrm.viewsets import ActivityViewSet


@pytest.mark.django_db
class TestSpecificFiltersActivities:
    def test_filter_clients_of(self, activity_factory, person_factory):
        person1 = person_factory()
        person2 = person_factory()
        activity_factory(participants=(person_factory(relationship_managers=(person1,)),))
        mvs = ActivityViewSet(kwargs={})
        request = APIRequestFactory().get("")
        request.user = UserFactory(is_active=True, is_superuser=True)
        mvs.request = APIRequestFactory().get("")
        qs = ActivityViewSet.get_model().objects.all()
        assert mvs.filterset_class(request=request).filter_clients_of(qs, "", None) == qs
        assert mvs.filterset_class(request=request).filter_clients_of(qs, "", person1).count() == 1
        assert mvs.filterset_class(request=request).filter_clients_of(qs, "", person2).count() == 0


@pytest.mark.django_db
class TestSpecificFiltersEntries:
    @pytest.mark.parametrize("base_class", [Person, Company])
    def test_choice_noactivity(self, activity_factory, base_class):
        person = PersonWithEmployerFactory()
        company = person.employer.first().employer
        entry_instance = person if base_class == Person else company
        activity = activity_factory(
            creator=person,
            assigned_to=person,
            latest_reviewer=person,
            preceded_by=None,
            participants=(person,),
            companies=(company,),
        )
        number_of_days_since_last_activity = (timezone.now() - activity.period.upper).days + 1
        number_of_days_no_activity = (timezone.now() - activity.period.upper).days - 1
        qs = base_class.objects.all()
        qs_count = qs.count()
        assert choice_noactivity(qs, "", None) == qs
        assert choice_noactivity(qs, "", 0).count() == qs_count - 1
        assert entry_instance not in choice_noactivity(qs, "", 0)
        assert choice_noactivity(qs, "", number_of_days_since_last_activity).count() == qs_count - 1
        assert entry_instance not in choice_noactivity(qs, "", number_of_days_since_last_activity)
        assert choice_noactivity(qs, "", number_of_days_no_activity).count() == qs_count



import pytest
from django.dispatch import receiver
from rest_framework import status
from rest_framework.test import APIRequestFactory
from termcolor import colored
from wbcore.contrib.directory.factories import PersonFactory
from wbcore.test import GenerateTest, default_config
from wbcore.test.mixins import TestViewSet
from wbcore.test.signals import custom_update_kwargs
from wbcore.test.utils import get_data_from_factory, get_kwargs, get_or_create_superuser

from wbcrm.factories import AccountFactory
from wbcrm.viewsets import ActivityParticipantModelViewSet
from wbcrm.viewsets.accounts import (
    ChildAccountAccountModelViewSet,
    InheritedAccountRoleAccountModelViewSet,
)

config = {}
for key, value in default_config.items():
    config[key] = list(
        filter(
            lambda x: x.__module__.startswith("wbcrm")
            and x.__name__
            not in [
                "ActivityParticipantModelViewSet",
                "InheritedAccountRoleAccountModelViewSet",
                "AccountRoleAccountModelViewSet",
                "ChildAccountAccountModelViewSet",
            ],
            value,
        )
    )


@pytest.mark.django_db
@GenerateTest(config)
class TestProject:
    pass


# ActivityParticipantModelViewSet Test
class ActivityParticipantTestViewSet(TestViewSet):
    def _get_mixins_update_data(self, type):
        api_request = APIRequestFactory()
        superuser = get_or_create_superuser()
        obj = self.factory()
        data = get_data_from_factory(obj, self.mvs, superuser=superuser, update=True)
        data["participant"] = PersonFactory().id
        if type == "PATCH":
            request = api_request.patch("", data)
        else:  # "UPDATE"
            request = api_request.put("", data)
        request.user = superuser
        kwargs = get_kwargs(obj, self.mvs, request=request, data=data)
        return obj, request, kwargs, data

    def test_patch_request(self):
        obj, request, kwargs, data = self._get_mixins_update_data("PATCH")
        vs = self.mvs.as_view({"patch": "partial_update"})
        ep = self._get_endpoint_config(request, kwargs, obj)
        ep_update = ep.get_instance_endpoint()
        response = vs(request, **kwargs, data=data)
        if ep_update:
            assert response.status_code == status.HTTP_200_OK, str(response.status_code) + f" == 200 ({response.data})"
            assert response.data.get("instance"), str(response.data.get("instance")) + " should not be empty"
        else:
            assert response.status_code == status.HTTP_405_METHOD_NOT_ALLOWED, (
                str(response.status_code) + f" == 405 ({response.data})"
            )
        print(f"- {self.__class__.__name__}:test_patchviewset", colored("PASSED", "green"))  # noqa: T201

    def test_update_request(self):
        obj, request, kwargs, _ = self._get_mixins_update_data("UPDATE")
        vs = self.mvs.as_view({"put": "update"})
        ep = self._get_endpoint_config(request, kwargs, obj)
        ep_update = ep.get_instance_endpoint()
        response = vs(request, **kwargs)
        if ep_update:
            assert response.status_code == status.HTTP_200_OK, str(response.status_code) + f" == 200 ({response.data})"
            assert response.data.get("instance"), str(response.data.get("instance")) + " should not be empty"
        else:
            assert response.status_code == status.HTTP_405_METHOD_NOT_ALLOWED, (
                str(response.status_code) + f" == 405 ({response.data})"
            )
        print(f"- {self.__class__.__name__}:test_update_request", colored("PASSED", "green"))  # noqa: T201


class GenerateActivityParticipantTest(GenerateTest):
    def test_modelviewsets(_self, mvs, client):
        my_test = ActivityParticipantTestViewSet(mvs)
        my_test.execute_test_list_endpoint(client)
        my_test.execute_test_detail_endpoint()


@pytest.mark.django_db
@GenerateActivityParticipantTest({"viewsets": [ActivityParticipantModelViewSet]})
class TestActivityParticipant:
    pass


@receiver(custom_update_kwargs, sender=ChildAccountAccountModelViewSet)
def receive_kwargs_child_account(sender, *args, **kwargs):
    if obj := kwargs.get("obj_factory"):
        parent = AccountFactory.create()
        obj.parent = parent
        obj.save()
        return {"account_id": parent.id}
    return {}


@receiver(custom_update_kwargs, sender=InheritedAccountRoleAccountModelViewSet)
def receive_kwargs_inherited_account_role(sender, *args, **kwargs):
    if obj := kwargs.get("obj_factory"):
        parent_account = AccountFactory.create()
        child_account = obj.account
        child_account.parent = parent_account
        child_account.save()
        obj.account = parent_account
        obj.save()
        return {"account_id": child_account.id}



from wbcore import serializers as wb_serializers
from wbcore.contrib.directory.serializers import EntryRepresentationSerializer

from ..models import Group


class GroupRepresentationSerializer(wb_serializers.RepresentationSerializer):
    endpoint = "wbcrm:grouprepresentation-list"
    _detail = wb_serializers.HyperlinkField(reverse_name="wbcrm:group-detail")

    class Meta:
        model = Group
        fields = (
            "id",
            "_detail",
            "title",
        )


class GroupModelSerializer(wb_serializers.ModelSerializer):
    _members = EntryRepresentationSerializer(source="members", many=True)

    class Meta:
        model = Group
        fields = (
            "id",
            "title",
            "members",
            "_members",
        )



from rest_framework.decorators import action
from wbcore import viewsets
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    create_simple_display,
)
from wbcore.viewsets import RepresentationViewSet

from wbcrm.filters import GroupFilter
from wbcrm.models import Group
from wbcrm.serializers import GroupModelSerializer, GroupRepresentationSerializer
from wbcrm.viewsets.display import GroupModelDisplay
from wbcrm.viewsets.endpoints import GroupEndpointConfig


class GroupRepresentationViewSet(RepresentationViewSet):
    queryset = Group.objects.all()
    search_fields = ("title",)
    serializer_class = GroupRepresentationSerializer


class GroupModelViewSet(viewsets.ModelViewSet):
    LIST_DOCUMENTATION = "wbcrm/markdown/documentation/group.md"
    IDENTIFIER = "wbcrm:group"
    ordering = ("title",)
    ordering_fields = ("title", "members")
    search_fields = ("title",)
    filterset_class = GroupFilter
    queryset = Group.objects.all().prefetch_related("members")
    serializer_class = GroupModelSerializer
    display_config_class = GroupModelDisplay
    endpoint_config_class = GroupEndpointConfig

    @action(methods=["GET"], detail=True)
    def pre_change(self, request, pk=None):
        return viewsets.PreAction(
            message="Do you really want to change this?",
            instance_display=create_simple_display([["title"]]),
        ).to_response()



from .accounts import ACCOUNT_MENUITEM
from .activities import (
    ACTIVITY_MENUTITEM,
    ACTIVITYCHART_MENUITEM,
    ACTIVITYTYPE_MENUITEM,
)
from .groups import GROUPS_MENUITEM
from .products import PRODUCT_MENUITEM



from django.utils.translation import gettext_lazy as _
from wbcore import filters as wb_filters
from wbcore.contrib.directory.models import Entry

from wbcrm.models import Group


class GroupFilter(wb_filters.FilterSet):
    members = wb_filters.ModelMultipleChoiceFilter(
        label=_("Members"),
        queryset=Entry.objects.all(),
        endpoint=Entry.get_representation_endpoint(),
        value_key=Entry.get_representation_value_key(),
        label_key=Entry.get_representation_label_key(),
    )

    class Meta:
        model = Group
        fields = {
            "title": ["exact", "icontains"],
        }



from typing import Dict, List, TypedDict


class _GoogleEntityInfo(TypedDict):
    id: str
    email: str
    displayName: str
    self: bool


class TimeInfo(TypedDict):
    date: str
    dateTime: str
    timeZone: str


class _AttendeeInfo(_GoogleEntityInfo):
    organizer: bool
    resource: bool
    optional: bool
    responseStatus: str
    comment: str
    additionalGuests: int


class _ExtendedPropertiesInfo(TypedDict):
    private: Dict[any, str]
    shared: Dict[any, str]


class _CreateRequestInfo(TypedDict):
    requestId: str
    conferenceSolutionKey: Dict[str, str]
    status: Dict[str, str]


class _ConferenceSolutionInfo(TypedDict):
    key: Dict[any, str]
    name: str
    iconUri: str


class _EntryPointInfo(TypedDict):
    entryPointType: str
    uri: str
    label: str
    pin: str
    accessCode: str
    meetingCode: str
    passcode: str
    password: str


class _ConferenceDataInfo(TypedDict):
    createRequest: _CreateRequestInfo
    entryPoints: List[_EntryPointInfo]
    conferenceSolution: _ConferenceSolutionInfo
    conferenceId: str
    signature: str
    notes: str


class _GadgetInfo(TypedDict):
    type: str
    title: str
    link: str
    iconLink: str
    width: int
    height: int
    display: str
    preferences: Dict[any, str]


class _OverridesInfo(TypedDict):
    method: str
    minutes: int


class _RemindersInfo(TypedDict):
    useDefault: bool
    overrides: List[_OverridesInfo]


class _URLInfo(TypedDict):
    url: str
    title: str


class _AttachmentInfo(TypedDict):
    fileUrl: str
    title: str
    mimeType: str
    iconLink: str
    fileId: str


class GoogleEventType(TypedDict):
    """
    The typing information for the Google event dictionary
    """

    kind: str
    id: str
    status: str
    htmlLink: str
    created: str
    updated: str
    summary: str
    description: str
    location: str
    colorId: str
    creator: _GoogleEntityInfo
    organizer: _GoogleEntityInfo
    start: TimeInfo
    end: TimeInfo
    endTimeUnspecified: bool
    recurrence: List[str]
    recurringEventId: str
    originalStartTime: TimeInfo
    transparency: str
    visibility: str
    iCalUID: str
    sequence: int
    attendees: List[_AttendeeInfo]
    attendeesOmitted: bool
    extendedProperties: _ExtendedPropertiesInfo
    hangoutLink: str
    conferenceData: _ConferenceDataInfo
    gadget: _GadgetInfo
    anyoneCanAddSelf: bool
    guestsCanInviteOthers: bool
    guestsCanModify: bool
    guestsCanSeeOtherGuests: bool
    privateCopy: bool
    locked: bool
    reminders: _RemindersInfo
    source: _URLInfo
    attachments: List[_AttachmentInfo]
    eventType: str



from django.db import models


class Event(models.Model):
    """
    we store the event notification we received from the webhook in this model to easily debug the sync
    """

    data = models.JSONField(default=dict, blank=True)
    result = models.JSONField(default=dict, blank=True)
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)



from datetime import timedelta

from django.utils import timezone
from django.utils.translation import gettext as _
from wbcore.menus import ItemPermission, MenuItem
from wbcore.permissions.shortcuts import is_internal_user


def default_activity_create_get_params(request) -> dict:
    return {"participants": request.user.profile.id}


def default_activitychart_get_params(request) -> dict:
    current_last_week_date_start = timezone.now().date() - timedelta(days=7)
    current_next_week_date_end = timezone.now().date() + timedelta(days=7)

    return {"period": f"{current_last_week_date_start:%Y-%m-%d},{current_next_week_date_end:%Y-%m-%d}"}


ACTIVITYTYPE_MENUITEM = MenuItem(
    label=_("Activity Types"),
    endpoint="wbcrm:activitytype-list",
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user),
        permissions=["wbcrm.view_activitytype"],
    ),
    add=MenuItem(
        label=_("Create Activity Type"),
        endpoint="wbcrm:activitytype-list",
        permission=ItemPermission(
            method=lambda request: is_internal_user(request.user),
            permissions=["wbcrm.add_activitytype"],
        ),
    ),
)


ACTIVITY_MENUTITEM = MenuItem(
    label=_("Activities"),
    endpoint="wbcrm:activity-list",
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user), permissions=["wbcrm.view_activity"]
    ),
    # endpoint_get_parameters={"only_recent": True}, # TODO confirm that this can actually be commented out
    add=MenuItem(
        label=_("Create Activity"),
        endpoint="wbcrm:activity-list",
        permission=ItemPermission(
            method=lambda request: is_internal_user(request.user), permissions=["wbcrm.add_activity"]
        ),
        endpoint_get_parameters=default_activity_create_get_params,
    ),
)
ACTIVITYCHART_MENUITEM = MenuItem(
    label=_("Activity Chart"),
    endpoint="wbcrm:activitychart-list",
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user), permissions=["wbcrm.view_activity"]
    ),
    endpoint_get_parameters=default_activitychart_get_params,
)



from wbcrm.viewsets.titles.activities import (
    ActivityChartModelTitleConfig,
    ActivityParticipantTitleConfig,
    ActivityTitleConfig,
    ActivityTypeTitleConfig,
)

from .accounts import (
    AccountRoleAccountTitleConfig,
    AccountTitleConfig,
    ChildAccountAccountTitleConfig,
)
from .products import ProductTitleConfig



from datetime import timedelta

from celery import shared_task
from dynamic_preferences.registries import global_preferences_registry
from googleapiclient.discovery import Resource
from psycopg.types.range import TimestamptzRange
from wbcore.contrib.authentication.models import User
from wbcrm.models import Activity

from ...typing_informations import GoogleEventType
from ...utils import GoogleSyncUtils
from .update import update_activity_participant


@shared_task(queue="synchronization")
def create_internal_activity_based_on_google_event(
    event: GoogleEventType, user: User, service: Resource, parent_occurrence: Activity | None = None, is_instance=False
):
    """
    A method for creating a "Workbench" activity based on a "Google" event. If the google event is a recurring event, this method will also create the corresponding workbench activities.

    :param event:           A google event body.
    :param user:            The current workbench user.
    :param service:         Thee google Resource.
    :param parent_occurrence: The parent activity. This is only used when the activity is part of a recurring chain. Per default it is None.
    :param is_instance:     True, when the event is a google instance (A instance is an event that is part of a recurring chain)
    """

    external_id = event["id"]
    if event.get("status") == "cancelled" or Activity.objects.filter(external_id=event["id"]).exists():
        return

    event_creator = event.get("organizer", {})
    event_creator_mail = event_creator.get("email", "")
    event_creator_displayed_name = event_creator.get("displayName", "")
    creator = GoogleSyncUtils.get_or_create_person(event_creator_mail, event_creator_displayed_name)
    event_start, event_end = GoogleSyncUtils.get_start_and_end(event)
    if event_start == event_end:
        event_end = event_end + timedelta(seconds=1)
    period = TimestamptzRange(event_start, event_end)  # type: ignore
    all_day: bool = True if event["start"].get("date") else False
    metadata = {"google_backend": {"instance": event}} if is_instance else {"google_backend": {"event": event}}

    act = Activity.objects.create(
        external_id=external_id,
        title=event.get("summary", "(No Subject)"),
        assigned_to=creator,
        creator=creator,
        description=event.get("description", ""),
        start=event_start,
        end=event_end,
        period=period,
        all_day=all_day,
        location=event.get("location"),
        visibility=GoogleSyncUtils.convert_event_visibility_to_activity_visibility(event.get("visibility", "")),
        metadata=metadata,
        parent_occurrence=parent_occurrence,
    )

    update_activity_participant(event, act)

    if event.get("recurrence") and not is_instance:
        instances: dict = service.events().instances(calendarId=user.email, eventId=event["id"]).execute()
        instance_items: list[GoogleEventType] = instances["items"]
        global_preferences = global_preferences_registry.manager()
        max_list_length: int = global_preferences["wbcrm__recurrence_maximum_count"]
        instance_items = instance_items[:max_list_length]

        for item in instance_items:
            if item["start"] == act.metadata["google_backend"].get("event", {}).get("start"):
                updated_metadata: dict = act.metadata
                updated_metadata["google_backend"] |= {"instance": item}
                Activity.objects.filter(id=act.id).update(metadata=updated_metadata)
            else:
                create_internal_activity_based_on_google_event(item, user, service, act, True)



from django.utils.translation import gettext as _
from wbcore.metadata.configs.titles import TitleViewConfig


class ProductTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return _("Products")

    def get_create_title(self):
        return _("Add Product")

    def get_instance_title(self):
        return _("Product")



# Generated by Django 5.0.9 on 2024-12-03 08:41

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbcrm", "0014_alter_account_relationship_status"),
    ]

    operations = [
        migrations.AlterField(
            model_name="activity",
            name="type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="activity",
                to="wbcrm.activitytype",
                verbose_name="Type",
            ),
        ),
    ]



from django.contrib import admin
from django.utils.translation import gettext_lazy as _
from reversion.admin import VersionAdmin

from wbcrm.models import Activity, ActivityParticipant, ActivityType
from wbcrm.models.activities import ActivityCompanyThroughModel


class ParticipantInline(admin.TabularInline):
    model = ActivityParticipant


class CompanyInline(admin.TabularInline):
    model = ActivityCompanyThroughModel


@admin.register(ActivityType)
class ActivityTypeAdmin(admin.ModelAdmin):
    search_fields = ("title",)
    list_display = ("id", "title")


class ActivityInline(admin.StackedInline):
    model = Activity
    fk_name = "parent_occurrence"
    extra = 0
    fieldsets = (
        (
            _("Main information"),
            {
                "fields": (
                    "title",
                    "description",
                )
            },
        ),
    )


@admin.register(Activity)
class ActivityAdmin(VersionAdmin):
    search_fields = ("title",)
    list_display = ("id", "status", "title", "period", "is_active", "parent_occurrence_id", "metadata")
    fieldsets = (
        (_("Main information"), {"fields": ("title", "description", "result", "creator")}),
        (_("Meta"), {"fields": ("status", "type", "visibility", "is_active", "metadata")}),
        (_("Temporal Information"), {"fields": ("period", "all_day")}),
        (_("Geographical Information"), {"fields": ("location", "location_longitude", "location_latitude")}),
        (_("Linked Entries"), {"fields": ("assigned_to", "groups")}),
        (_("Linked Activities"), {"fields": ("preceded_by",)}),
        (
            _("Recurrence"),
            {
                "fields": (
                    "repeat_choice",
                    "parent_occurrence",
                    ("recurrence_count", "recurrence_end"),
                    "propagate_for_all_children",
                )
            },
        ),
    )

    raw_id_fields = (
        "assigned_to",
        "participants",
        "groups",
        "preceded_by",
        "creator",
        "latest_reviewer",
        "parent_occurrence",
    )

    inlines = [ActivityInline, ParticipantInline, CompanyInline]

    def reversion_register(self, model, **options):
        options = {
            "exclude": (
                "created",
                "creator",
                "edited",
            ),
        }
        super().reversion_register(model, **options)

    def get_queryset(self, request):
        return Activity.all_objects.all()


@admin.register(ActivityParticipant)
class ActivityParticipantAdmin(VersionAdmin):
    search_fields = ("activity", "participant")
    list_display = ("id", "activity", "participant", "participation_status")

    def reversion_register(self, model, **options):
        options = {
            "follow": (
                "activity",
                "participant",
                "participation_status",
            )
        }
        super().reversion_register(model, **options)



import pytest
from wbcore.contrib.directory.typings import Person as PersonDTO

from wbcrm.typings import Activity as ActivityDTO
from wbcrm.typings import ParticipantStatus as ParticipantStatusDTO
from wbcrm.typings import User as UserDTO


@pytest.mark.django_db
class TestDTO:
    @pytest.mark.parametrize("user_id", [None, 1])
    def test_user_comparison(self, user_id):
        user1 = UserDTO(1, {})
        user2 = UserDTO(id=user_id, metadata={})
        if user1.id == user2.id is not None:
            assert user1 == user2
        else:
            assert user1 != user2
        assert user2 is not None

    @pytest.mark.parametrize(
        "id, name, email",
        [
            (None, None, None),
            (None, "test", None),
            (None, "test", "test@test.com"),
            (1, None, None),
            (1, "test", None),
            (1, "test", "test@test.com"),
        ],
    )
    def test_participation_status(self, id, name, email):
        person1 = PersonDTO(1, "test", "test", "test@test.com")
        person2 = PersonDTO(id=id, first_name=name, last_name=name, email=email)

        participant1 = ParticipantStatusDTO(person=person1)
        participant2 = ParticipantStatusDTO(person=person2)
        if (participant1.id == participant2.id is not None) or (
            participant1.person == participant2.person and participant1.activity == participant2.activity
        ):
            assert participant1 == participant2
        else:
            assert participant1 != participant2
        assert participant2 is not None

    @pytest.mark.parametrize(
        "id, title, email",
        [
            (None, None, None),
            (None, "test", None),
            (None, "test", "test@test.com"),
            (1, None, None),
            (1, "test", None),
            (1, "test", "test@test.com"),
        ],
    )
    def test_activity(self, id, title, email):
        activity1 = ActivityDTO(id=1, title="test", metadata={})
        activity2 = ActivityDTO(id=id, title=title, metadata={})
        if activity1.id == activity2.id is not None:
            assert activity1 == activity2
        else:
            assert activity1 != activity2
        assert activity2 is not None



# Generated by Django 5.0.9 on 2024-11-19 14:11

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbcrm", "0013_account_action_plan_account_relationship_status_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="account",
            name="relationship_status",
            field=models.PositiveIntegerField(
                blank=True,
                help_text="The Relationship Status from 1 to 5. 1 being the cold and 5 being the hot.",
                null=True,
                validators=[django.core.validators.MinValueValidator(1), django.core.validators.MaxValueValidator(5)],
                verbose_name="Relationship Status",
            ),
        ),
    ]



import json
import re
import time
from datetime import date, timedelta
from typing import Any

from dateutil import parser
from dateutil.relativedelta import relativedelta
from django.contrib.auth import get_user_model
from django.http import HttpRequest, HttpResponse
from django.utils import timezone
from django.utils.translation import gettext, gettext_lazy
from dynamic_preferences.registries import global_preferences_registry
from psycopg.types.range import TimestamptzRange

from wbcrm.synchronization.activity.backend import SyncBackend
from wbcrm.synchronization.activity.utils import flattened_dict_into_nested_dict
from wbcrm.typings import Activity as ActivityDTO
from wbcrm.typings import ParticipantStatus as ParticipantStatusDTO
from wbcrm.typings import User as UserDTO

from .msgraph import MicrosoftGraphAPI
from .parser import OutlookParser, parse

User = get_user_model()


class OutlookSyncBackend(SyncBackend):
    METADATA_KEY = "outlook"

    def open(self):
        self.msgraph = MicrosoftGraphAPI()

    def close(self):
        pass

    def _validation_response(self, request: HttpRequest) -> HttpResponse:
        # handle validation
        if token := request.GET.get("validationToken"):
            return HttpResponse(token, content_type="text/plain")
        # handle callback consent permision
        if _consent := request.GET.get("admin_consent"):
            return HttpResponse(_consent, content_type="text/plain")
        return super()._validation_response(request)

    def _is_inbound_request_valid(self, request: HttpRequest) -> bool:
        is_valid = False
        try:
            if request.body and (json_body := json.loads(request.body)) and (notifications := json_body.get("value")):
                # When many changes occur, MSGraph may send multiple notifications that correspond to different subscriptions in the same POST request.
                for notification in parse(notifications):
                    if (
                        notification.get("client_state")
                        == global_preferences_registry.manager()["wbactivity_sync__outlook_sync_client_state"]
                        and (resource_data := notification.get("resource_data"))
                        and (event_type := resource_data.get("@odata.type", "").lower())
                        and (event_type == "#microsoft.graph.event")
                    ):
                        is_valid = True
                    else:
                        return False
        except json.decoder.JSONDecodeError:
            return False
        return is_valid

    def _get_events_from_request(self, request: HttpRequest) -> list[dict[str, Any]]:
        events = []
        for notification in parse(json.loads(request.body)["value"]):
            event = {
                "change_type": notification["change_type"],
                "resource": notification["resource"],
                "subscription_id": notification["subscription_id"],
            }
            if data := self.msgraph.get_event_by_resource(notification["resource"]):
                # Try to get the organizer's event in case if it's an invitation event
                if data["is_organizer"] is True:
                    event["organizer_resource"] = notification["resource"]
                elif (
                    (tenant_id := self.msgraph.get_tenant_id(data["organizer.email_address.address"]))
                    and data["is_organizer"] is False
                    and (organizer_data := self.msgraph.get_event_by_uid(tenant_id, data["uid"]))
                ):
                    event["organizer_resource"] = f"Users/{tenant_id}/Events/{organizer_data['id']}"
                    data = organizer_data
                event.update(data)
            events.append(event)
        return events

    def _deserialize(self, event: dict[str, Any], include_metadata: bool = True) -> tuple[ActivityDTO, bool, UserDTO]:
        event_id = event.get("id")
        delete_notification = event.get("change_type") == "deleted"
        is_deleted = delete_notification or not event_id or event.get("is_cancelled") is True
        user_dto = UserDTO(metadata={self.METADATA_KEY: {"subscription": {"id": event.get("subscription_id")}}})
        if include_metadata:
            _metadata = {"event_uid": event["uid"], "event_id": event_id} if event_id else {}
            if resource := event.get("resource"):
                _metadata["resources"] = [resource]
            if organizer_resource := event.get("organizer_resource"):
                _metadata["organizer_resource"] = organizer_resource
            elif is_deleted and resource:
                _metadata["organizer_resource"] = resource
            metadata = {self.METADATA_KEY: _metadata}
        else:
            metadata = {}

        if event_id:
            start = OutlookParser.convert_string_to_datetime(event["start.date_time"], event["start.time_zone"])
            end = OutlookParser.convert_string_to_datetime(event["end.date_time"], event["end.time_zone"])
            if start == end:
                end += timedelta(seconds=1)
            period = TimestamptzRange(start, end)
            if event["is_all_day"] is True:
                period = OutlookParser.convert_to_all_day_period(period)
            participants, conference_room = OutlookParser.deserialize_participants(event)
            recurrence_dict = OutlookParser.deserialize_recurring_activities(event)
            if event.get("is_reminder_on"):
                reminder_choice = OutlookParser.convert_reminder_minutes_to_choice(
                    event.get("reminder_minutes_before_start")
                )
            else:
                reminder_choice = ActivityDTO.ReminderChoice.NEVER.name
            activity_dto = ActivityDTO(
                metadata=metadata,
                title=event["subject"] if event.get("subject") else "(No Subject)",
                period=period,
                description=event.get("body.content", ""),
                participants=participants,
                creator=OutlookParser.deserialize_person(
                    event["organizer.email_address.address"], event["organizer.email_address.name"]
                ),
                visibility=OutlookParser.convert_sensitivity_to_visibility(event.get("sensitivity")),
                reminder_choice=reminder_choice,
                is_cancelled=event.get("is_cancelled", False),
                all_day=event.get("is_all_day", False),
                online_meeting=event.get("is_online_meeting", False),
                location=event.get("location.display_name"),
                conference_room=conference_room[0] if conference_room else None,
                recurrence_end=recurrence_dict["recurrence_end"],
                recurrence_count=recurrence_dict["recurrence_count"],
                repeat_choice=recurrence_dict["repeat_choice"],
            )
            if (
                event["type"] == "seriesMaster"
                and (tenant_id := self.msgraph.get_tenant_id(activity_dto.creator.email))
                and (
                    occurrences := self.msgraph.get_instances_event(
                        tenant_id, event["id"], start.date(), end.date() + relativedelta(years=10)
                    )
                )
            ):
                recurring_activities = []
                for occurrence in occurrences:
                    instance_dto, *_ = self._deserialize(occurrence, include_metadata=False)
                    instance_dto.recurrence_end = activity_dto.recurrence_end
                    instance_dto.recurrence_count = activity_dto.recurrence_count
                    instance_dto.repeat_choice = activity_dto.repeat_choice
                    instance_dto.metadata = {
                        self.METADATA_KEY: {
                            **metadata[self.METADATA_KEY],
                            "occurrence_id": occurrence["id"],
                            "occurrence_resource": f"Users/{tenant_id}/Events/{occurrence['id']}",
                        }
                    }
                    if (extension := self.msgraph.get_extension_event(tenant_id, occurrence["id"])) and (
                        act_id := extension.get("activity_id")
                    ):
                        instance_dto.id = act_id
                    recurring_activities.append(instance_dto)
                activity_dto.recurring_activities = recurring_activities
        else:
            activity_dto = ActivityDTO(metadata=metadata, title=event.get("subject", "(No Subject)"))
        activity_dto.delete_notification = delete_notification
        return activity_dto, is_deleted, user_dto

    def _serialize(self, activity_dto: ActivityDTO, created: bool = False) -> dict[str, Any]:
        attendees = OutlookParser.serialize_participants(activity_dto.participants)
        if activity_dto.conference_room:
            attendees.append(OutlookParser.serialize_conference_room(activity_dto.conference_room))

        activity_dict = {
            "subject": activity_dto.title,
            "start": {
                "dateTime": activity_dto.period.lower.isoformat(),
                "timeZone": activity_dto.period.lower.tzname(),
            },
            "end": {
                "dateTime": activity_dto.period.upper.isoformat(),
                "timeZone": activity_dto.period.upper.tzname(),
            },
            "body": {
                "contentType": "HTML",
                "content": activity_dto.description,
            },
            "attendees": attendees,
            "sensitivity": OutlookParser.convert_visibility_to_sensitivity(activity_dto.visibility),
            "reminderMinutesBeforeStart": OutlookParser.convert_reminder_choice_to_minutes(
                activity_dto.reminder_choice
            ),
            "isReminderOn": False if activity_dto.reminder_choice == ActivityDTO.ReminderChoice.NEVER.name else True,
            "isAllDay": activity_dto.all_day,
            "responseRequested": True,
        }

        if activity_dto.online_meeting:
            activity_dict.update(
                {"allowNewTimeProposals": True, "isOnlineMeeting": True, "onlineMeetingProvider": "teamsForBusiness"}
            )
        if activity_dto.location:
            activity_dict["location"] = {
                "displayName": (
                    activity_dto.location if isinstance(activity_dto.location, str) else str(activity_dto.location)
                )
            }
        activity_dict["locations"] = [{"displayName": activity_dto.location}] if activity_dto.location else []
        if created and activity_dto.is_recurrent:
            activity_dict.update(OutlookParser.serialize_recurring_activities(activity_dto))
        return activity_dict

    def _stream_deletion(self, activity_dto: ActivityDTO):
        if event := self.get_external_event(activity_dto):
            if tenant_id := self.msgraph.get_tenant_id(activity_dto.creator.email):
                self.msgraph.delete_event(tenant_id, event["id"])

    def _stream_creation(
        self, activity_dto: ActivityDTO, activity_dict: dict[str, Any]
    ) -> list[tuple[ActivityDTO, dict[str, Any]]] | None:
        if not self.get_external_event(activity_dto) and (
            tenant_id := self.msgraph.get_tenant_id(activity_dto.creator.email)
        ):
            if event := self.msgraph.create_event(tenant_id, activity_dict):
                metadata_list = self._get_metadata_from_event(activity_dto, event)
                for act_dto, _metadata in metadata_list:
                    if act_dto.is_recurrent:
                        self._stream_extension_event(act_dto, _metadata)
                return metadata_list

    def _stream_update(
        self,
        activity_dto: ActivityDTO,
        activity_dict: dict[str, Any] | None = None,
        only_participants_changed: bool = False,
        external_participants: list = [],
        keep_external_description: bool = False,
    ) -> tuple[ActivityDTO, dict[str, Any]]:
        if old_event := self.get_external_event(activity_dto):
            if tenant_id := self.msgraph.get_tenant_id(activity_dto.creator.email):
                activity_dict = activity_dict or old_event
                # When a participant is removed, we send a complete update of the event to remove this person in outlook
                activity_dict["attendees"] += external_participants
                if only_participants_changed:
                    # An event update that includes only the attendees property in the request body sends a meeting update to only the attendees that have changed.
                    # An event update that removes an attendee specified as a member of a distribution list sends a meeting update to all the attendees.
                    activity_dict = {"attendees": activity_dict["attendees"]}
                elif keep_external_description:
                    # preserve body when description is not synchronized especially meeting blob for online meeting.
                    activity_dict["body"]["content"] = old_event.get("body.content", "")

                if (
                    activity_dto.is_recurrent
                    and activity_dto.propagate_for_all_children
                    and not activity_dto.is_leaf
                    and (old_master_event := self.get_external_event(activity_dto, master_event=True))
                ):
                    if activity_dto.is_root:
                        self.msgraph.update_event(tenant_id, old_master_event["id"], activity_dict)
                    else:
                        return self._stream_update_from_occurrence(activity_dto, old_master_event, activity_dict)
                else:
                    self.msgraph.update_event(tenant_id, old_event["id"], activity_dict)
        elif not activity_dto.metadata.get(self.METADATA_KEY):
            # creation of recurring activity is done only during the creation phase of the activity
            if activity_dto.is_recurrent and activity_dict.get("recurrence"):
                activity_dict.pop("recurrence")
            return self._stream_creation(activity_dto, activity_dict)

    def _stream_update_from_occurrence(self, activity_dto: ActivityDTO, old_master_event: dict, activity_dict: dict):
        if (tenant_id := self.msgraph.get_tenant_id(activity_dto.creator.email)) and (
            instances_dict := self._get_occurrences_from_master_event(
                old_master_event, start_date_as_key=True, from_date=activity_dto.period.lower.date()
            )
        ):
            # Update recurrence pattern from old root to new root
            remaining_old_recurrence = {
                "recurrence": flattened_dict_into_nested_dict(OutlookParser.serialize_event_keys(old_master_event))[
                    "recurrence"
                ]
            }
            last_start_old = activity_dto.period.lower - timedelta(days=1)
            if remaining_old_recurrence := OutlookParser.serialize_recurrence_range_with_end_date(
                remaining_old_recurrence, last_start_old
            ):
                self.msgraph.update_event(tenant_id, old_master_event["id"], remaining_old_recurrence)
            # Create recurrence from new root
            new_recurrence = OutlookParser.serialize_recurring_activities(activity_dto)
            last_start_new = sorted(instances_dict.keys()).pop()
            if new_recurrence := OutlookParser.serialize_recurrence_range_with_end_date(
                new_recurrence, last_start_new
            ):
                activity_dict.update(new_recurrence)
                if new_master_event := self.msgraph.create_event(tenant_id, activity_dict):
                    # delete occurrences that had been deleted in the past
                    if new_instances_dict := self._get_occurrences_from_master_event(
                        new_master_event, start_date_as_key=True
                    ):
                        for key, value in new_instances_dict.items():
                            if not instances_dict.get(key):
                                self.msgraph.delete_event(tenant_id, value["id"])
                    metadata_list = self._get_metadata_from_event(activity_dto, new_master_event)
                    for act_dto, _metadata in metadata_list:
                        if act_dto.is_recurrent:
                            self._stream_extension_event(act_dto, _metadata)
                    return metadata_list

    def _stream_update_only_attendees(self, activity_dto: ActivityDTO, participants_dto: list[ParticipantStatusDTO]):
        new_attendees = OutlookParser.serialize_participants(participants_dto)
        self._stream_update(
            activity_dto=activity_dto,
            external_participants=new_attendees,
            only_participants_changed=True,
        )

    def _stream_forward(self, activity_dto: ActivityDTO, participants_dto: list[ParticipantStatusDTO]):
        if (event := self.get_external_event(activity_dto)) and (
            tenant_id := self.msgraph.get_tenant_id(activity_dto.creator.email)
        ):
            new_attendees = []
            participants, _ = OutlookParser.deserialize_participants(event)
            external_emails_participants = [participant.person.email for participant in participants]
            for participant_dto in participants_dto:
                # forward if it is not the organizer or if it is not already part of the participants
                if (
                    participant_dto.person.email != activity_dto.creator.email
                    and participant_dto.person.email not in external_emails_participants
                ):
                    new_attendees.append(OutlookParser.serialize_person(participant_dto.person))
                    self.msgraph.forward_event(tenant_id, event["id"], new_attendees)

    def _stream_participant_change(
        self, participant_dto: ParticipantStatusDTO, is_deleted: bool = False, wait_before_changing: bool = False
    ):
        if (organizer_event := self.get_external_event(participant_dto.activity)) and (
            tenant_id := self.msgraph.get_tenant_id(participant_dto.person.email)
        ):
            if wait_before_changing:
                time.sleep(10)
            if invitation_event := self.msgraph.get_event_by_uid(tenant_id, organizer_event["uid"]):
                if participant_dto.person.email == participant_dto.activity.creator.email:
                    if is_deleted or participant_dto.status == ParticipantStatusDTO.ParticipationStatus.CANCELLED.name:
                        pass  # self.msgraph.cancel_event(tenant_id, invitation_event["id"]) # delete activity if organizer is removed of participants
                else:
                    if is_deleted or participant_dto.status == ParticipantStatusDTO.ParticipationStatus.CANCELLED.name:
                        self.msgraph.decline_event(tenant_id, invitation_event["id"])
                    elif participant_dto.status == ParticipantStatusDTO.ParticipationStatus.MAYBE.name:
                        self.msgraph.tentatively_accept_event(tenant_id, invitation_event["id"])
                    elif participant_dto.status in [
                        ParticipantStatusDTO.ParticipationStatus.ATTENDS.name,
                        ParticipantStatusDTO.ParticipationStatus.ATTENDS_DIGITALLY.name,
                    ]:
                        self.msgraph.accept_event(tenant_id, invitation_event["id"])

    def _stream_extension_event(self, activity_dto: ActivityDTO, metadata: dict | None = None) -> None:
        metadata = metadata if metadata else activity_dto.metadata
        if (
            (tenant_id := self.msgraph.get_tenant_id(activity_dto.creator.email))
            and (_metadata := metadata.get(self.METADATA_KEY))
            and (occurrence_id := _metadata.get("occurrence_id"))
        ):
            self.msgraph.update_or_create_extension_event(tenant_id, occurrence_id, {"activity_id": activity_dto.id})

    def _set_web_hook(self, user: "User") -> dict[str, dict[str, Any]]:
        response = {}
        self.open()
        if (
            (outlook := user.metadata.get(self.METADATA_KEY))
            and (subscription := outlook.get("subscription"))
            and (ms_subscription := self.msgraph.subscription(subscription.get("id")))
        ):
            response = self.msgraph._renew_subscription(ms_subscription.get("id"))
        elif tenant_id := self.msgraph.get_tenant_id(user.email):
            response = self.msgraph._subscribe(f"users/{tenant_id}/events/", "created, updated, deleted")
        else:
            raise Exception(gettext("Outlook TenantId not found for: ") + str(user))
        self.close()
        return {"subscription": response} if response else user.metadata.get(self.METADATA_KEY, {})

    def _stop_web_hook(self, user: "User") -> dict[str, dict[str, Any]]:
        if (
            (outlook_backend := user.metadata.get(self.METADATA_KEY))
            and (subscription := outlook_backend.get("subscription"))
            and (subscription_id := subscription.get("id"))
        ):
            self.open()
            self.msgraph._unsubscribe(subscription_id)
            self.close()
            del user.metadata[self.METADATA_KEY]["subscription"]
        else:
            raise ValueError(str(user) + gettext(" has no active webhook"))
        return user.metadata.get(self.METADATA_KEY)

    def check_web_hook(self, user: "User", raise_error: bool = True) -> bool:
        error = ""
        self.open()
        if (outlook := user.metadata.get(self.METADATA_KEY)) and (subscription := outlook.get("subscription")):
            if not self.msgraph.subscription(subscription.get("id")):
                error = gettext("Webhook is invalid, Remove or Stop it and Set again please. ")
        else:
            error = gettext("Webhook not found. ")
        if error and raise_error:
            if user_tenant_id := self.msgraph.get_tenant_id(user.email):
                tenant_ids = [
                    items[1]
                    for sub in self.msgraph.subscriptions()
                    if (resource := sub.get("resource")) and (items := resource.split("/")) and len(items) > 2
                ]
                error += gettext(
                    "Number of subscriptions found in outlook for {} out of the total number: {}/{}."
                ).format(user.email, tenant_ids.count(user_tenant_id), len(tenant_ids))
            else:
                error += gettext("TenantId not found for ") + str(user)
            raise Exception(error)
        self.close()
        return False if error else True

    def renew_web_hooks(self) -> None:
        self.open()
        lookup = {f"metadata__{self.METADATA_KEY}__subscription__id__isnull": False}
        for user in User.objects.filter(**lookup):
            if response := self.msgraph._renew_subscription(user.metadata[self.METADATA_KEY]["subscription"]["id"]):
                user.metadata[self.METADATA_KEY]["subscription"] = response
                user.save()
        self.close()

    def _get_webhook_inconsistencies(self) -> str:
        self.open()
        subscriptions = self.msgraph.subscriptions()
        calls = set(
            list(
                map(
                    lambda x: x["id"],
                    list(filter(lambda x: x["resource"] == "/communications/callRecords", subscriptions)),
                )
            )
        )
        calendars = set(
            list(
                map(
                    lambda x: x["id"],
                    list(filter(lambda x: bool(re.match(r"users\/.*\/events\/", x["resource"])), subscriptions)),
                )
            )
        )
        lookup = f"metadata__{self.METADATA_KEY}__subscription__id"
        stored_calendars = set(User.objects.filter(**{f"{lookup}__isnull": False}).values_list(lookup, flat=True))
        message = ""
        if len(calls) == 0:
            message += gettext_lazy("<li>No Call Record subscription found in Microsoft</li>")
        if calendars != stored_calendars:
            diff = calendars.difference(stored_calendars)
            diff_inv = stored_calendars.difference(calendars)
            message += gettext_lazy(
                """
                <li>Number of calendar subscription not found in our system : <b>{}</b></li>
                <p>{}</p>

                <li>Number of calendar subscription assumed to be active not found in outlook: <b>{}</b></li>
                <p>{}</p>
            """
            ).format(len(diff), diff, len(diff_inv), diff_inv)
        if message == "":
            for subscription in subscriptions:
                if expiration_date_time := subscription.get("expiration_date_time"):
                    date_time = parser.parse(expiration_date_time)
                    diff = date_time - timezone.now()
                    if diff.days < 1:
                        message += f"Resource: {subscription['resource']} expires on {date_time} (in {diff})"
        self.close()
        return message

    def get_external_event(self, activity_dto: ActivityDTO, master_event: bool = False) -> dict:
        event = None
        metadata = activity_dto.metadata.get(self.METADATA_KEY, {})
        if master_event or not activity_dto.is_recurrent:
            if resource := metadata.get("organizer_resource"):
                event = self.msgraph.get_event_by_resource(resource)
            if (creator := activity_dto.creator) and (tenant_id := self.msgraph.get_tenant_id(creator.email)):
                if not event and (event_id := metadata.get("event_id")):
                    event = self.msgraph.get_event(tenant_id, event_id)
                if not event and (event_uid := metadata.get("event_uid")):
                    event = self.msgraph.get_event_by_uid(tenant_id, event_uid)
        else:
            if resource := metadata.get("occurrence_resource"):
                event = self.msgraph.get_event_by_resource(resource)
            elif (
                not event
                and (creator := activity_dto.creator)
                and (tenant_id := self.msgraph.get_tenant_id(creator.email))
                and metadata.get("occurrence_id")
            ):
                event = self.msgraph.get_event(tenant_id, metadata["occurrence_id"])
        return event

    def get_external_participants(
        self, activity_dto: ActivityDTO, internal_participants_dto: list[ParticipantStatusDTO]
    ) -> list:
        external_participants = []
        if event := self.get_external_event(activity_dto):
            internal_emails = [
                participant_dto.person.email for participant_dto in internal_participants_dto if participant_dto.person
            ]
            for participant in event.get("attendees", []):
                if participant["emailAddress"]["address"] not in internal_emails:
                    external_participants.append(participant)
        return external_participants

    def _is_participant_valid(self, user: "User") -> bool:
        try:
            return super()._is_participant_valid(user) and self.check_web_hook(user, raise_error=False)
        except Exception:
            return False

    def _generate_event_metadata(self, tenant_id: str, master_event: dict, occurrence_event: dict = {}):
        resource = f"Users/{tenant_id}/Events/{master_event['id']}"
        metadata = {
            self.METADATA_KEY: {
                "resources": [resource],
                "organizer_resource": resource,
                "event_uid": master_event["uid"],
                "event_id": master_event["id"],
            }
        }
        if occurrence_event:
            metadata[self.METADATA_KEY].update(
                {
                    "occurrence_id": occurrence_event["id"],
                    "occurrence_resource": f"Users/{tenant_id}/Events/{occurrence_event['id']}",
                }
            )
        return metadata

    def _get_metadata_from_event(
        self, activity_dto: ActivityDTO, event: dict
    ) -> list[tuple[ActivityDTO, dict[str, Any]]]:
        metadata_list = []
        if tenant_id := self.msgraph.get_tenant_id(activity_dto.creator.email):
            if event.get("type") == "seriesMaster" and (
                instances_dict := self._get_occurrences_from_master_event(event, start_date_as_key=True)
            ):
                occurrence = instances_dict.get(activity_dto.period.lower.date(), {})
                metadata_list.append((activity_dto, self._generate_event_metadata(tenant_id, event, occurrence)))
                instances_dto = activity_dto.recurring_activities + activity_dto.invalid_recurring_activities
                for instance_dto in instances_dto:
                    if occurrence := instances_dict.get(instance_dto.period.lower.date()):
                        metadata_list.append(
                            (instance_dto, self._generate_event_metadata(tenant_id, event, occurrence))
                        )
            else:
                metadata_list.append((activity_dto, self._generate_event_metadata(tenant_id, event)))
        return metadata_list

    def _get_occurrences_from_master_event(
        self, event: dict, start_date_as_key: bool = False, from_date: date = None
    ) -> dict:
        """
        Dict of instances event whose key is id by default otherwise start date and the value is the occurrence event data.
        """
        start = (
            from_date
            if from_date
            else OutlookParser.convert_string_to_datetime(event["start.date_time"], event["start.time_zone"]).date()
        )
        end = OutlookParser.convert_string_to_datetime(
            event["end.date_time"], event["end.time_zone"]
        ).date() + relativedelta(years=10)
        tenant_id = self.msgraph.get_tenant_id(event["organizer.email_address.address"])
        occurrences_dict = {}
        if occurrences := self.msgraph.get_instances_event(tenant_id, event["id"], start, end):
            if start_date_as_key:
                occurrences_dict = {
                    OutlookParser.convert_string_to_datetime(
                        occurrence["start.date_time"], occurrence["start.time_zone"]
                    ).date(): occurrence
                    for occurrence in occurrences
                }
            else:
                occurrences_dict = {occurrence["id"]: occurrence for occurrence in occurrences}
        return occurrences_dict



from typing import Optional

from django.utils.translation import gettext as _
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display import Display
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    create_simple_display,
)
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class GroupModelDisplay(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=(
                dp.Field(key="title", label=_("Name")),
                dp.Field(key="members", label=_("Members")),
            ),
        )

    def get_instance_display(self) -> Display:
        return create_simple_display([["title"], ["members"]])



from .accounts import AccountRoleModelAdmin, AccountModelAdmin
from .activities import ActivityAdmin, ActivityTypeAdmin
from .groups import GroupModelAdmin
from .products import ProductAdmin
from .events import EventModelAdmin



from __future__ import absolute_import, unicode_literals

import logging
from datetime import date, datetime, time, timedelta

from celery import shared_task
from django.contrib.auth import get_user_model
from django.db.models import (
    F,
    FloatField,
    Func,
    Max,
    OuterRef,
    Q,
    QuerySet,
    Subquery,
    Sum,
    Value,
)
from django.db.models.functions import Cast, Coalesce, Least
from django.template.loader import render_to_string
from django.utils import timezone
from django.utils.timezone import make_aware
from django.utils.translation import gettext as _
from dynamic_preferences.registries import global_preferences_registry
from psycopg.types.range import TimestamptzRange
from rest_framework.reverse import reverse
from wbcore.contrib.directory.models import Company, Person
from wbcore.contrib.notifications.dispatch import send_notification

from wbcrm.models import Activity, ActivityType

logger = logging.getLogger()
User = get_user_model()


@shared_task
def notify(time_offset: int = 60, now: datetime | None = None):
    """
    Cron task that runs every 60s and checks which activities will happen during the notify interval

    Arguments:
        time_offset (int):  The notification period. Defaults to 60.
        now (datetime | None, optional):  The time at which activity needs to be checked for a notification. Defaults to None.
    """

    if not now:
        now = timezone.now()
    base_queryset: QuerySet[Activity] = Activity.objects.filter(Q(status=Activity.Status.PLANNED))
    reminder_choices = Activity.ReminderChoice.values

    # we don't notify activity if Reminder is Never
    reminder_choices.remove(Activity.ReminderChoice.NEVER)
    for reminder in reminder_choices:
        # get the reminder correspondance in minutes
        reminder_minutes = Activity.ReminderChoice.get_minutes_correspondance(reminder)
        reminder_range = TimestamptzRange(
            now + timedelta(minutes=reminder_minutes),
            now + timedelta(minutes=reminder_minutes) + timedelta(seconds=time_offset),
        )  # type: ignore #ErrMsg: Expected no arguments to "TimestamptzRange" constructor
        # get all incoming activity with same reminder that happen during the notify interval
        upcoming_occurence = base_queryset.filter(
            reminder_choice=reminder, period__startswith__contained_by=reminder_range
        )
        for activity in upcoming_occurence:
            participants = activity.get_participants()

            # For each Employee in the activity participants
            for employee in Person.objects.filter_only_internal().filter(id__in=participants.values("id")).all():
                # formant and create Notification
                activity_type_label = activity.type.title
                desc = (
                    activity.description
                    if activity.description and activity.description not in ["", "<p></p>"]
                    else None
                )
                message = render_to_string(
                    "email/activity.html",
                    {
                        "participants": participants,
                        "type": activity_type_label,
                        "title": activity.title,
                        "start": activity.period.lower,
                        "end": activity.period.upper,
                        "description": desc,
                    },
                )
                send_notification(
                    code="wbcrm.activity.reminder",
                    title=_("{type} in {reminder} Minutes").format(
                        type=activity_type_label, reminder=reminder_minutes
                    ),
                    body=message,
                    user=employee.user_account,
                    reverse_name="wbcrm:activity-detail",
                    reverse_args=[activity.pk],
                )


@shared_task
def yesterdays_activity_summary(yesterday: date | None = None, report_receiver_user_ids: list[int] | None = None):
    """A daily task that sends a summary of all employees' yesterday's activities to the users assigned to this task

    Args:
        yesterday (date | None, optional): Date of the previous day. Defaults to None.
    """

    if not yesterday:
        yesterday = date.today() - timedelta(days=1)
    yesterday = datetime.combine(yesterday, time(0, 0, 0))  # we convert the date to datetime
    time_range = TimestamptzRange(make_aware(yesterday), make_aware(yesterday + timedelta(days=1)))  # type: ignore #ErrMsg: Expected no arguments to "TimestamptzRange" constructor

    # Create the list of all employees' activities for yesterday
    employees_list: list[Person] = list(Person.objects.filter_only_internal())
    internal_activities: QuerySet[Activity] = (
        Activity.objects.exclude(status=Activity.Status.CANCELLED)
        .filter(period__overlap=time_range, participants__in=employees_list)
        .order_by("period__startswith")
    )

    if not (internal_activities.exists() or report_receiver_user_ids):
        return

    activity_lists: list[list[dict]] = []  # contains an activity list for each employee
    employee_names = []
    for employee in employees_list:
        if internal_activities.filter(participants=employee).exists():
            employees_activities = internal_activities.filter(participants=employee)
            # Create activity list with formatted activity dictionaries for employee
            activity_lists.append(
                [
                    {
                        "type": activity.type.title,
                        "title": activity.title,
                        "start": activity.period.lower,  # type: ignore #ErrMsg: Cannot access member "lower" for type "TimestamptzRange"
                        "end": activity.period.upper,  # type: ignore #ErrMsg: Cannot access member "upper" for type "TimestamptzRange"
                        "endpoint": reverse("wbcrm:activity-detail", args=[activity.pk]),
                    }
                    for activity in employees_activities
                ]
            )
            employee_names.append(employee.full_name)

    # Create the notification for each person with the right permission
    for user in User.objects.filter(id__in=report_receiver_user_ids).distinct():
        context = {
            "map_activities": zip(employee_names, activity_lists),
            "activities_count": internal_activities.count(),
            "report_date": yesterday.strftime("%d.%m.%Y"),
        }
        message = render_to_string("email/global_daily_summary.html", context)
        send_notification(
            code="wbcrm.activity.daily_summary",
            title=_("Activity Summary {}").format(yesterday.strftime("%d.%m.%Y")),
            body=message,
            user=user,
        )


@shared_task
def todays_activity_summary(today: date | None = None):
    """Creates a summary of the daily upcoming activities for all employees

    Args:
        today (date | None, optional): Date of today. Defaults to None.
    """

    if not today:
        today = date.today()
    tz_info = timezone.get_current_timezone()
    today_range = TimestamptzRange(
        lower=datetime.combine(today, time(0, 0, 0)).replace(tzinfo=tz_info),
        upper=datetime.combine(today + timedelta(days=1), time(0, 0, 0).replace(tzinfo=tz_info)),
    )

    for employee in Person.objects.filter_only_internal():
        # Get all the employee's activities from that day
        activity_qs: QuerySet[Activity] = (
            Activity.objects.exclude(status=Activity.Status.CANCELLED)
            .filter(period__overlap=today_range, participants=employee)
            .order_by("period__startswith")
        )

        # Create the formatted activity dictionaries
        activity_list = []
        for activity in activity_qs:
            activity_list.append(
                {
                    "type": activity.type.title,
                    "title": activity.title,
                    "start": activity.period.lower,
                    "end": activity.period.upper,
                    "endpoint": reverse("wbcrm:activity-detail", args=[activity.pk]),
                }
            )

        # Create the proper notification for each employee
        if activity_list and employee.primary_email_contact():
            context = {"activities": activity_list}
            message = render_to_string("email/daily_summary.html", context)
            send_notification(
                code="wbcrm.activity.daily_summary",
                title=_("Your Schedule for Today"),
                body=message,
                user=employee.user_account,
            )


@shared_task
def finish(now: datetime | None = None):
    """Cron task running every X Seconds. Checks all activities that have finished and sends a reminder to review the activity to the assigned person.

    Args:
        now (datetime | None, optional): Current datetime. Defaults to None.
    """

    if not now:
        now = timezone.now()
    # Get all finished activities during the cron task interval
    finished_activities: QuerySet[Activity] = Activity.objects.filter(
        Q(status=Activity.Status.PLANNED.name)
        & Q(repeat_choice=Activity.ReoccuranceChoice.NEVER)
        & Q(period__endswith__lte=now)
    )

    # For each of these activities, Send the Notification to the person in charge of that activity
    for activity in finished_activities:
        activity.finish()
        activity.save()
        if (assignee := activity.assigned_to) and assignee.is_internal and assignee.user_account:
            send_notification(
                code="wbcrm.activity.finished",
                title=_("Activity Finished"),
                body=_('The activity "{title}" just finished and you are in charge of it. Please review.').format(
                    title=activity.title
                ),
                user=assignee.user_account,
                reverse_name="wbcrm:activity-detail",
                reverse_args=[activity.pk],
            )


@shared_task
def default_activity_heat_calculation(check_datetime: datetime | None = None):
    """A script that calculates the activity heat of companies and
    persons on a scale from 0.0 to 1.0. The type and time interval of
    completed activities serve as the basis for the heat calculation for companies. A person's rating is based
    on the score of the person's employer.

    Args:
        check_datetime (datetime | None, optional): The datetime of the activity heat check. Defaults to None.
    """

    if not check_datetime:
        check_datetime = timezone.now()

    class JulianDay(Func):
        """
        The Julian day is the continuous count of days since the beginning of the Julian period.
        """

        function = ""
        output_field = FloatField()  # type: ignore #ErrMsg: Expression of type "FloatField[float]" cannot be assigned to declared type "property"

        def as_postgresql(self, compiler, connection):
            self.template = "CAST (to_char(%(expressions)s, 'J') AS INTEGER)"
            return self.as_sql(compiler, connection)

    global_preferences_manager = global_preferences_registry.manager()
    main_company: int = global_preferences_manager["directory__main_company"]

    external_employees: QuerySet[Person] = Person.objects.exclude(id__in=Person.objects.filter_only_internal())
    external_companies: QuerySet[Company] = Company.objects.exclude(id=main_company)
    # Calculate the activity heat of a person in the last 180 days.
    activity_score = (
        Activity.objects.filter(
            companies__id=OuterRef("id"),
            status__in=["REVIEWED", "FINISHED"],
            period__endswith__gte=check_datetime - timedelta(days=180),
            period__endswith__lte=check_datetime,
        )
        .annotate(
            ratio=JulianDay(Value(check_datetime)) - JulianDay(F("period__endswith")),
            date_score=(Value(365) - F("ratio")) / Value(365.0),
            score=Cast(F("type__score"), FloatField()) * F("date_score"),
        )
        .values("companies__id")
        .annotate(sum_score=Sum(F("score")))
        .values("sum_score")
    )

    company_score = (
        Company.objects.filter(
            id=OuterRef("id"),
        )
        .annotate(
            norm_score=Coalesce(
                Subquery(activity_score) / float(ActivityType.Score.MAX),
                Value(0.0),
            ),
            abs_norm_score=Least(F("norm_score"), 1.0),
        )
        .values("abs_norm_score")
    )
    employer_max_score = (
        external_companies.filter(employees__id=OuterRef("id"))
        .values("employees__id")
        .annotate(max_score=Max("activity_heat"))
        .values("max_score")
    )

    external_companies.update(activity_heat=Subquery(company_score))
    external_employees.filter(employers__id__in=external_companies).update(activity_heat=Subquery(employer_max_score))



import pytest
from rest_framework import status
from rest_framework.test import APIRequestFactory
from wbcore.test.utils import get_kwargs

from wbcrm.factories import ActivityFactory
from wbcrm.viewsets import ActivityChartModelViewSet, ActivityViewSet


@pytest.mark.django_db
class TestSpecificsChartViewsets:
    @pytest.mark.parametrize(
        "mvs, factory",
        [
            (ActivityChartModelViewSet, ActivityFactory),
        ],
    )
    def test_option_request(self, mvs, factory, superuser):
        request = APIRequestFactory().options("")
        request.user = superuser
        factory()
        kwargs = {"user_id": request.user.id}
        vs = mvs.as_view({"options": "options"})
        response = vs(request, **kwargs)
        assert response.status_code == status.HTTP_200_OK
        assert response.data

    @pytest.mark.parametrize(
        "mvs, factory",
        [
            (ActivityChartModelViewSet, ActivityFactory),
        ],
    )
    def test_viewsets(self, mvs, factory, superuser):
        request = APIRequestFactory().get("")
        request.user = superuser
        factory()
        kwargs = {"user_id": request.user.id}
        vs = mvs.as_view({"get": "list"})
        response = vs(request, **kwargs)
        assert response.status_code == status.HTTP_200_OK
        assert response.data


@pytest.mark.django_db
class TestSpecificsInfiniteViewsets:
    @pytest.mark.parametrize(
        "mvs, factory",
        [
            (
                ActivityViewSet,
                ActivityFactory,
            ),
        ],
    )
    def test_option_request(self, mvs, factory, superuser):
        request = APIRequestFactory().options("")
        request.user = superuser
        obj = factory(participants=(request.user.profile,))
        request.GET = request.GET.copy()
        request.GET["date_gte"] = str(obj.start.date())
        kwargs = {}
        mvs.request = request
        vs = mvs.as_view({"options": "options"})
        kwargs = get_kwargs(obj, mvs, request=request)
        response = vs(request, **kwargs).render()
        assert response.status_code == status.HTTP_200_OK
        assert response.data

    @pytest.mark.parametrize(
        "mvs, factory",
        [
            (
                ActivityViewSet,
                ActivityFactory,
            ),
        ],
    )
    def test_viewsets(self, mvs, factory, superuser):
        request = APIRequestFactory().get("")
        request.user = superuser
        obj = factory(participants=(request.user.profile,))
        request.GET = request.GET.copy()
        request.GET["date_gte"] = str(obj.start.date())
        kwargs = {}
        mvs.request = request
        vs = mvs.as_view({"get": "list"})
        response = vs(request, **kwargs).render()
        assert response.status_code == status.HTTP_200_OK
        assert response.data

    @pytest.mark.parametrize(
        "mvs, factory",
        [
            (
                ActivityViewSet,
                ActivityFactory,
            ),
        ],
    )
    def test_viewsets_without_date_gte(self, mvs, factory, superuser):
        request = APIRequestFactory().get("")
        request.user = superuser
        factory(participants=(request.user.profile,))
        request.GET = request.GET.copy()
        request.GET["date_gte"] = None
        kwargs = {}
        mvs.request = request
        vs = mvs.as_view({"get": "list"})
        response = vs(request, **kwargs).render()
        assert response.status_code == status.HTTP_200_OK
        assert response.data



import copy
import dataclasses
import operator
from functools import reduce
from typing import Any

from django.conf import settings
from django.db import transaction
from django.db.models import Q, QuerySet
from django.http import HttpRequest, HttpResponse
from django.utils import timezone
from wbcore.contrib.agenda.models import Building, ConferenceRoom
from wbcore.contrib.directory.models import EmailContact, Person

from wbcrm.models import Activity, ActivityParticipant, ActivityType, Event
from wbcrm.synchronization.activity.utils import flattened_metadata_lookup
from wbcrm.typings import Activity as ActivityDTO
from wbcrm.typings import ConferenceRoom as ConferenceRoomDTO
from wbcrm.typings import ParticipantStatus as ParticipantStatusDTO
from wbcrm.typings import Person as PersonDTO
from wbcrm.typings import User as UserDTO

from .backend import SyncBackend
from .preferences import (
    can_sync_cancelled_activity,
    can_sync_cancelled_external_activity,
    can_sync_create_new_activity_on_replanned_reviewed_activity,
    can_sync_past_activity,
    can_synchronize_activity_description,
    can_synchronize_external_participants,
)


class ActivityController:
    update_fields = [
        "title",
        "period",
        "visibility",
        "creator",
        "conference_room",
        "reminder_choice",
        "is_cancelled",
        "all_day",
        "online_meeting",
        "location",
        "description",
    ]

    def __init__(self, backend: SyncBackend):
        if settings.DEBUG:
            raise ValueError("DEBUG Mode -> activity sync disabled")
        self.backend = backend()

    def _is_valid(self, activity_dto: ActivityDTO) -> bool:
        """
        Check if the activity can be synchronized or not.
        - Past activity is not synchronized if global preference is False
        - Activity is synchronized if at least one of the participants has an active webhook
        """
        if period := activity_dto.period:
            if period.upper < timezone.now() and not can_sync_past_activity():
                return False
            else:
                return self.backend.is_valid(activity_dto)
        else:
            qs_activities = self.get_activities(activity_dto)
            with transaction.atomic():
                return qs_activities.exists() and any(
                    [self._is_valid(act._build_dto()) for act in qs_activities if act.period]
                )

    def handle_inbound_validation_response(self, request: HttpRequest) -> HttpResponse:
        """
        allows to send a response to the external calendar if it is required before receiving the events in the webhook
        """
        return self.backend._validation_response(request)

    def get_events_from_inbound_request(self, request: HttpRequest) -> list[dict[str, Any]]:
        """
        allows to get list of event following the notification
        """
        events = []
        self.backend.open()
        if self.backend._is_inbound_request_valid(request):
            for event in self.backend._get_events_from_request(request):
                events.append(event)
        self.backend.close()
        return events

    def handle_inbound(self, event: dict, event_object_id: int) -> None:
        """
        Each event received in the webhook is processed here, it allows to delete or create/update the corresponding activity
        """
        self.backend.open()
        activity_dto, is_deleted, user_dto = self.backend._deserialize(event)
        if activity_dto.is_recurrent or self._is_valid(activity_dto):
            if is_deleted:
                self.delete_activity(activity_dto, user_dto, event_object_id)
            else:
                self.update_or_create_activity(activity_dto, event_object_id)
        self.backend.close()

    def _handle_outbound_data_preferences(self, activity_dto: ActivityDTO) -> tuple[ActivityDTO, list]:
        """
        Activity data is parsed to take into account the company's preferences
        this allows you to exclude the description or exclude external participants from the data if global preference is True
        """
        if not can_synchronize_activity_description():
            activity_dto.description = ""
        valid_participants = []
        internal_participants = []
        for participant_dto in activity_dto.participants:
            # External person are removed of the list if global preference is True
            if (
                Person.all_objects.get(id=participant_dto.person.id).is_internal
                or can_synchronize_external_participants()
            ):
                internal_participants.append(participant_dto)
                if participant_dto.status != ActivityParticipant.ParticipationStatus.CANCELLED:
                    valid_participants.append(participant_dto)

        activity_dto.participants = valid_participants

        # list of external participants present in the external event, they will be added to the participants before updating the event to avoid deleting them
        external_participants = (
            self.backend.get_external_participants(activity_dto, internal_participants)
            if not can_synchronize_external_participants()
            else []
        )
        return activity_dto, external_participants

    def handle_outbound(
        self, activity_dto: ActivityDTO, old_activity_dto: ActivityDTO = None, is_deleted: bool = False
    ):
        """
        Requests sent to the external calendar are processed here
        It allows to send requests to delete, modify or create the event in the external calendar corresponding to the activity
        """
        if not self._is_valid(activity_dto):
            return
        self.backend.open()
        if is_deleted or activity_dto.is_cancelled:
            self.backend._stream_deletion(activity_dto)
        else:
            activities_metadata = []
            created = True if not old_activity_dto else False
            # dataclasses.replace returns a new copy of the object without passing in any changes, return a copy with no modification
            activity_dto_preference, external_participants = self._handle_outbound_data_preferences(
                dataclasses.replace(activity_dto)
            )
            activity_dict = self.backend._serialize(activity_dto_preference, created=created)
            if created:  # then it's a creation
                activities_metadata = self.backend._stream_creation(activity_dto, activity_dict)
            elif self._has_changed(activity_dto, old_activity_dto):
                keep_external_description = not can_synchronize_activity_description()
                only_participants_changed = self._has_changed(
                    activity_dto, old_activity_dto, update_fields=["participants"]
                ) and not self._has_changed(activity_dto, old_activity_dto, exclude_fields=["participants"])
                activities_metadata = self.backend._stream_update(
                    activity_dto,
                    activity_dict,
                    only_participants_changed,
                    external_participants,
                    keep_external_description,
                )

            if activities_metadata:
                for act_dto, act_metadata in activities_metadata:
                    self.update_activity_metadata(act_dto, act_metadata)
        self.backend.close()

    def handle_outbound_participant(
        self,
        participant_dto: ParticipantStatusDTO,
        old_participant_dto: ParticipantStatusDTO = None,
        is_deleted: bool = False,
    ):
        """
        allows to update the status of the event in the external calendar to match the one updated in the internal activity
        """
        # check if activity creator is internal or activity is not passed according to global preference
        if not self._is_valid(participant_dto.activity):
            return

        # check if participant is internal or can sync external participant is allowed according to global preference
        if not (
            Person.all_objects.get(id=participant_dto.person.id).is_internal or can_synchronize_external_participants()
        ):
            return

        self.backend.open()
        was_cancelled = False
        status_changed = False
        if old_participant_dto and old_participant_dto.status != participant_dto.status:
            was_cancelled = old_participant_dto.status == ActivityParticipant.ParticipationStatus.CANCELLED
            status_changed = participant_dto.status not in [
                ActivityParticipant.ParticipationStatus.NOTRESPONDED,
                ActivityParticipant.ParticipationStatus.PENDING_INVITATION,
            ]

        wait_before_changing_status = False
        if (not is_deleted and not old_participant_dto) or was_cancelled:
            # self.backend._stream_forward(participant_dto.activity, [participant_dto])
            # forward event doesn't work when user is an external, following the documentation it's recommended to only include the attendees property in the request body
            # It will only send notifications to newly added attendees.
            self.backend._stream_update_only_attendees(
                activity_dto=participant_dto.activity, participants_dto=[participant_dto]
            )
            wait_before_changing_status = status_changed
        if is_deleted or status_changed:
            self.backend._stream_participant_change(
                participant_dto, is_deleted, wait_before_changing=wait_before_changing_status
            )
        self.backend.close()

    def handle_outbound_external_participants(self, activity_dto, participants_dto: list[ParticipantStatusDTO]):
        """
        allows to update the status of the event in the external calendar to match the one updated in the internal activity
        """
        # check if activity creator is internal or activity is not passed according to global preference
        if not self._is_valid(activity_dto):
            return

        self.backend.open()
        self.backend._stream_update_only_attendees(activity_dto=activity_dto, participants_dto=participants_dto)
        self.backend.close()

    def _changed_participants(
        self, participants: list[ParticipantStatusDTO], old_participants: list[ParticipantStatusDTO]
    ) -> bool:
        """
        Comparison of 2 lists of participants, returns false if they are identical
        """
        d1 = {elt.person.email: elt.status for elt in participants}
        d2 = {elt.person.email: elt.status for elt in old_participants}
        if set(d1.keys()) == set(d2.keys()):
            return any([d1[key] != d2[key] for key in d1.keys()])
        return True

    def _has_changed(
        self,
        activity_dto: ActivityDTO,
        old_activity_dto: ActivityDTO,
        update_fields: list = [],
        exclude_fields: list = [],
    ) -> bool:
        """
        Comparison of 2 activities, returns false if they are identical

        :param update_fields: allows to specify the list of fields taken into account in the comparison,
        if not specified we use the list of the controller

        :param exclude_fields: allows you to exclude fields from the comparison list
        """
        if not can_synchronize_activity_description():
            exclude_fields.append("description")
        update_fields = (
            update_fields
            if update_fields
            else (
                self.update_fields + ["propagate_for_all_children", "exclude_from_propagation"]
                if self.update_fields
                else activity_dto.__dataclass_fields__
            )
        )
        fields = list(set(update_fields) - set(exclude_fields))
        if "participants" in fields:
            fields.remove("participants")
            participants_changed = self._changed_participants(activity_dto.participants, old_activity_dto.participants)
        else:
            participants_changed = False
        is_new_activity = True if activity_dto and not old_activity_dto else False
        return (
            participants_changed
            or is_new_activity
            or (
                activity_dto
                and old_activity_dto
                and any(
                    [getattr(activity_dto, field, None) != getattr(old_activity_dto, field, None) for field in fields]
                )
            )
        )

    def get_activities(self, activity_dto: ActivityDTO, _operator: operator = operator.or_) -> QuerySet["Activity"]:
        """
        Received events are deserialized into data transfer object of activity,
        we use the metadata construct to identify the activity
        the operator allows to know which operation combination to perform during the filter
        """
        if conditions := [
            Q(**{key: value}) for key, value in flattened_metadata_lookup(activity_dto.metadata, key_string="metadata")
        ]:
            return Activity.all_objects.select_for_update().filter(reduce(_operator, conditions))
        return Activity.objects.none()

    def get_activity_participant(self, user_dto: UserDTO) -> Person:
        """
        Attendees of the external event are deserialized into person data transfer objects
        we use the metadata construct to identify the person
        """
        if conditions := [
            Q(**{key: value})
            for key, value in flattened_metadata_lookup(user_dto.metadata, key_string="user_account__metadata")
        ]:
            try:
                return Person.objects.get(reduce(operator.and_, conditions))
            except Exception:
                return None
        return None

    def _get_data_from_activity_dto(self, activity_dto: ActivityDTO, parent_occurrence: Activity = None) -> dict:
        """
        Data transfer object of activity obtained from the external event is parsed into a dict to allow the creation or update of the object in the database
        """
        activity_data = {}
        fields = self.update_fields if self.update_fields else activity_dto.__dataclass_fields__
        if not can_synchronize_activity_description() and "description" in fields:
            fields.remove("description")

        for field in fields:
            activity_data[field] = getattr(activity_dto, field)
        if activity_data.get("creator"):
            activity_data["creator"] = self.get_or_create_person(activity_dto.creator)
        if activity_data.get("conference_room"):
            activity_data["conference_room"] = self.get_or_create_conference_room(activity_dto.conference_room)
        if activity_dto.is_recurrent:
            activity_data.update(
                {
                    "recurrence_end": activity_dto.recurrence_end,
                    "recurrence_count": activity_dto.recurrence_count,
                    "repeat_choice": activity_dto.repeat_choice,
                }
            )
            if parent_occurrence:
                activity_data["parent_occurrence"] = parent_occurrence
        return activity_data

    def _create_activity(self, activity_dto: ActivityDTO, parent_occurrence: Activity = None) -> Activity:
        """
        Uses the Data transfer object obtained from the external event to create the activity in the database
        """
        activity_data = self._get_data_from_activity_dto(activity_dto, parent_occurrence)
        activity_type, created = ActivityType.objects.get_or_create(
            slugify_title="meeting", defaults={"title": "Meeting"}
        )
        activity = Activity(**activity_data, type=activity_type)
        activity.save(synchronize=False)
        if activity.period.lower < activity.created:
            # A created past event should not appear at the top of the list.
            Activity.objects.filter(id=activity.id).update(created=activity.period.lower, edited=activity.period.lower)
        self.update_or_create_participants(activity, activity_dto.participants)
        self.update_activity_metadata(activity._build_dto(), activity_dto.metadata)
        return activity

    def _update_activity(
        self, activity: Activity, activity_dto: ActivityDTO, parent_occurrence: Activity = None
    ) -> tuple[Activity, str]:
        """
        Convert the data transfer object obtained from the external event into a dict to update the activity in the database.
        """
        activity_data = self._get_data_from_activity_dto(activity_dto, parent_occurrence)
        if has_changed := self._has_changed(activity_dto, activity._build_dto(), update_fields=activity_data.keys()):
            # When previously canceled by the only internal participant is finally accepted by the latter
            if (
                (creator := activity.creator)
                and not creator.is_internal
                and activity.status == Activity.Status.CANCELLED
            ):
                activity_data["status"] = Activity.Status.PLANNED
            Activity.objects.filter(id=activity.id).update(**activity_data)
            activity.refresh_from_db()
            activity.save(synchronize=False)
        _, participant_changed = self.update_or_create_participants(activity, activity_dto.participants)
        self.update_activity_metadata(activity._build_dto(), activity_dto.metadata)
        return activity, f"activity_changes: {has_changed}, participants_changes: {participant_changed or False}; "

    @transaction.atomic
    def update_or_create_activity(self, activity_dto: ActivityDTO, event_object_id: int | None = None) -> None:
        """
        allows you to create or update a single or recurring activity from a data transfer object obtained from an external event
        """
        qs_activities = self.get_activities(activity_dto)
        if activity_dto.is_recurrent:
            event_result = {"action_type": "update or create recurring activities", "action": ""}
            ids_dto_dict = {}
            dates_dto_dict = {}
            for _dto in activity_dto.recurring_activities:
                if _dto.id:
                    ids_dto_dict[_dto.id] = _dto
                dates_dto_dict[_dto.period.lower.date()] = _dto
            ids_dto_list = set(ids_dto_dict.keys())
            dates_dto_list = set(dates_dto_dict.keys())

            for act in qs_activities.exclude(Q(id__in=ids_dto_list) | Q(period__startswith__date__in=dates_dto_list)):
                act.delete(synchronize=False)
            for act in qs_activities.filter(Q(id__in=ids_dto_list) | Q(period__startswith__date__in=dates_dto_list)):
                act_dto = (
                    _act_dto if (_act_dto := ids_dto_dict.get(act.id)) else dates_dto_dict[act.period.lower.date()]
                )
                activity_updated, _ = self._update_activity(act, act_dto)
                if not act_dto.id:
                    self.backend._stream_extension_event(activity_updated._build_dto())

            dates_act_dict = {act.period.lower.date(): act for act in qs_activities}
            if dates_to_create := dates_dto_list.difference(set(dates_act_dict.keys())):
                parent_start = sorted(dates_dto_list)[0]
                if dates_act_dict.get(parent_start):
                    parent_act = dates_act_dict[parent_start]
                else:
                    parent_dto = dates_dto_dict[parent_start]
                    if parent_dto.id and (instance := Activity.all_objects.filter(id=parent_dto.id).first()):
                        Activity.all_objects.filter(id=parent_dto.id).update(parent_occurrence=None, is_active=True)
                        parent_act, _ = self._update_activity(instance, parent_dto)
                    else:
                        if parent_act := self._create_activity(parent_dto):
                            qs_activities.exclude(id=parent_act.id).update(parent_occurrence=parent_act)
                            self.backend._stream_extension_event(parent_act._build_dto())
                if parent_start in dates_to_create:
                    dates_to_create.remove(parent_start)
                for _start_date in sorted(dates_to_create):
                    instance_dto = dates_dto_dict[_start_date]
                    if (instance_dto.id and (instance := Activity.all_objects.filter(id=instance_dto.id).first())) or (
                        instance := parent_act.child_activities.filter(period__startswith__date=_start_date).first()
                    ):
                        new_activity, _ = self._update_activity(instance, instance_dto, parent_occurrence=parent_act)
                    else:
                        new_activity = self._create_activity(instance_dto, parent_occurrence=parent_act)
                        self.backend._stream_extension_event(new_activity._build_dto())
        else:
            if qs_activities.exists():
                event_result = {"action_type": "update", "action": ""}
                for activity in qs_activities:
                    previous_status = activity.status
                    if not (
                        previous_status == Activity.Status.REVIEWED
                        and can_sync_create_new_activity_on_replanned_reviewed_activity()
                        and activity._build_dto().period != activity_dto.period
                    ):
                        activity_updated, participant_updates_info = self._update_activity(activity, activity_dto)
                        msg = f"{activity.id} => {previous_status} activity, {participant_updates_info} -> {activity_updated.status} activity; "
                        event_result["action"] += msg
                    else:
                        new_activity = self._create_activity(activity_dto)
                        event_result["action"] += (
                            f"reviewed activity {activity.id} replanned to {new_activity.id} -> new activity {new_activity.status} created"
                        )
                        # remove metadata since the old activity is not sync anymore
                        Activity.objects.filter(id=activity.id).update(
                            metadata={
                                self.backend.METADATA_KEY: {
                                    "info": f"reviewed activity has been replanned to {new_activity.id}"
                                }
                            }
                        )
            else:
                new_activity = self._create_activity(activity_dto)
                event_result = {
                    "action_type": "create",
                    "action": f"{new_activity.id} -> new activity {new_activity.status} created",
                }
        Event.objects.filter(id=event_object_id).update(result=event_result)

    def update_activity_metadata(self, activity_dto: ActivityDTO, new_metadata):
        """
        allows to update the metadata used to save the external event id which is used to retrieve the event/activity
        """
        old_metadata = activity_dto.metadata.get(self.backend.METADATA_KEY, {})
        metadata = copy.deepcopy(old_metadata)
        new_metadata = new_metadata.get(self.backend.METADATA_KEY, {})
        for key, new_value in new_metadata.items():
            if old_value := metadata.get(key):
                if isinstance(old_value, list):
                    values = set(new_value) if isinstance(new_value, list) else {new_value}
                    if new_values := [_value for _value in values if _value not in old_value]:
                        metadata[key] = sorted(old_value + new_values)
                else:
                    metadata[key] = new_value
            else:
                metadata[key] = new_value
        if old_metadata != metadata:
            activity_dto.metadata[self.backend.METADATA_KEY] = metadata
            Activity.objects.filter(id=activity_dto.id).update(metadata=activity_dto.metadata)

    def _cancel_or_delete_activity(self, activity: "Activity") -> None:
        """
        when an event is deleted in the external calendar, we cancel or delete the activity according to global preferences
        """
        if can_sync_cancelled_activity():
            if activity.status != Activity.Status.CANCELLED:
                activity.cancel()
                activity.save(synchronize=False)
        else:
            activity.delete(synchronize=False)

    def _delete_activity(self, activity: Activity, participant: Person, exact_match: bool):
        previous_status = activity.status
        internal_activity = activity.creator.is_internal if activity.creator else False
        is_internal_creator = internal_activity and activity.creator == participant
        if exact_match and is_internal_creator:
            if activity.parent_occurrence:
                # delete instance activity if it's not the parent of recurring activities
                activity.delete(synchronize=False)
            else:
                # single activity or first parent of recurring activity
                self._cancel_or_delete_activity(activity)
            msg = f"{activity.id} => {previous_status} activity {activity.status}; "
        else:
            activity.activity_participants.filter(participant=participant).update(
                participation_status=ActivityParticipant.ParticipationStatus.CANCELLED
            )
            msg = f"{activity.id} => participant status cancelled; "

        # Handle external creator activity cancellation if no internal participants exist
        internal_participants_exist = (
            activity.activity_participants.exclude(
                participation_status=ActivityParticipant.ParticipationStatus.CANCELLED
            )
            .exclude(participant=participant)
            .filter(participant__in=Person.objects.filter_only_internal())
            .exists()
        )
        if not internal_activity and can_sync_cancelled_external_activity() and not internal_participants_exist:
            self._cancel_or_delete_activity(activity)
            msg += f"external {previous_status} activity {activity.status}; "

        return msg

    @transaction.atomic
    def delete_activity(
        self, activity_dto: ActivityDTO, user_dto: UserDTO, event_object_id: int | None = None
    ) -> None:
        event_result = {"action_type": "delete", "action": ""}

        # Try to get the participant, return if not found
        participant = self.get_activity_participant(user_dto)
        if not participant:
            Event.objects.filter(id=event_object_id).update(result=event_result)
            return

        # Prepare query sets for activities deletion
        qs_activities = self.get_activities(activity_dto, operator.and_)  # activities deleted by the organizer
        qs_invitation_activities = self.get_activities(activity_dto)  # activities deleted by a participant

        # skip deletion when notification received is not a deletion
        if not activity_dto.delete_notification:
            activities_ids = list(qs_invitation_activities.values_list("id", "status"))
            event_result["action"] += f"{activities_ids} => skip deletion; wait for delete notification"
            Event.objects.filter(id=event_object_id).update(result=event_result)
            return

        # Handle activities deleted by organizer
        if qs_activities.exists():
            for activity in qs_activities:
                event_result["action"] += self._delete_activity(activity, participant, exact_match=True)

        # Handle activities with participant invitation
        elif qs_invitation_activities.exists():
            for activity in qs_invitation_activities:
                event_result["action"] += self._delete_activity(activity, participant, exact_match=False)

        event_result["participant"] = participant.id
        Event.objects.filter(id=event_object_id).update(result=event_result)

    def get_or_create_person(self, person_dto: PersonDTO) -> Person:
        """
        A method to get or create the internal person of a "External" event. Returns a Person objects of the internal database.

        :param mail: person mail.
        """
        potential_persons = Person.objects.filter(emails__address=person_dto.email.lower()).order_by(
            "-emails__primary"
        )
        if potential_persons.exists():
            person = potential_persons.first()  # TODO change with owner feature
        else:
            person = Person.objects.create(
                last_name=person_dto.last_name, first_name=person_dto.first_name, is_draft_entry=True
            )
            if (
                potential_contact := EmailContact.objects.filter(entry__isnull=True, address=person_dto.email.lower())
                .order_by("-primary")
                .first()
            ):
                EmailContact.objects.filter(id=potential_contact.id).update(entry=person)
            else:
                EmailContact.objects.create(entry=person, address=person_dto.email, primary=True)
        return person

    def get_or_create_conference_room(self, conference_room: ConferenceRoomDTO) -> ConferenceRoom:
        if ConferenceRoom.objects.filter(email=conference_room.email).exists():
            conference_room = ConferenceRoom.objects.get(email=conference_room.email)
        else:
            name_building = conference_room.name_building if conference_room.name_building else conference_room.email
            name_conference_room = conference_room.name if conference_room.name else conference_room.email
            building, _ = Building.objects.get_or_create(name=name_building)
            conference_room = ConferenceRoom.objects.create(
                name=name_conference_room, email=conference_room.email, building=building
            )
        return conference_room

    def update_or_create_participants(
        self, activity: Activity, participants_dto: list[ParticipantStatusDTO]
    ) -> tuple[list[ActivityParticipant], str]:
        """
        allows to create or update the status of the participants of an activity
        """
        activity_participants = []
        participants_changed = ""
        for participant_dto in participants_dto:
            person = self.get_or_create_person(participant_dto.person)
            kwargs = {"participation_status": participant_dto.status} if participant_dto.status else {}

            if activity_participant := ActivityParticipant.objects.filter(
                activity=activity, participant=person
            ).first():
                if activity_participant.participation_status != participant_dto.status:
                    ActivityParticipant.objects.filter(activity=activity, participant=person).update(**kwargs)
                    participants_changed += (
                        f"- {person}: {activity_participant.participation_status} -> {participant_dto.status}"
                    )
            else:
                activity_participant = ActivityParticipant.objects.create(
                    activity=activity, participant=person, **kwargs
                )
                participants_changed += f"- new participant {person}: {participant_dto.status}"
            activity_participants.append(activity_participant)
        return activity_participants, participants_changed



from dynamic_preferences.registries import global_preferences_registry


def can_sync_past_activity() -> bool:
    return global_preferences_registry.manager()["wbactivity_sync__sync_past_activity"]


def can_sync_cancelled_activity() -> bool:
    return global_preferences_registry.manager()["wbactivity_sync__sync_cancelled_activity"]


def can_sync_cancelled_external_activity() -> bool:
    return global_preferences_registry.manager()["wbactivity_sync__sync_cancelled_external_activity"]


def can_sync_create_new_activity_on_replanned_reviewed_activity() -> bool:
    return global_preferences_registry.manager()[
        "wbactivity_sync__sync_create_new_activity_on_replanned_reviewed_activity"
    ]


def can_synchronize_activity_description() -> bool:
    return global_preferences_registry.manager()["wbactivity_sync__sync_activity_description"]


def can_synchronize_external_participants() -> bool:
    return global_preferences_registry.manager()["wbactivity_sync__sync_external_participants"]
